"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the hashing helpers.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from typing import Optional, Union

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes

from PyFunceble.helpers.file import FileHelper


class HashHelper:
    """
    Simplify the hashing of data or file content.

    :param str algo:
        The algorithm to use for hashing.

    :raise ValueError: When the given algo is not known.
    """

    _algo: str = "SHA512_224"

    def __init__(self, algo: Optional[str] = None):
        if algo is not None:
            self.algo = algo

    @property
    def algo(self) -> str:
        """
        Provides the current state fo the :code:`_algo` attribute.
        """

        return self._algo

    @algo.setter
    def algo(self, value: str) -> None:
        """
        Sets the algorithm to work with.

        :param value:
            The name of the hash to use.

        :raise TypeError:
            When :code:`value` is not a :py:class:`str`.
        """

        if not isinstance(value, str):
            raise TypeError(f"<value> should be {str}, {type(value)} given.")

        value = value.upper()

        if not hasattr(hashes, value):
            raise ValueError(
                f"<value> ({value!r}) in an unknown algorithm ({self.algo!r})."
            )

        self._algo = value

    def set_algo(self, value: str) -> "HashHelper":
        """
        Sets the algorithm to work with.

        :param value:
            The name of the hash to use.
        """

        self.algo = value

        return self

    def __get_hash(self) -> hashes.Hash:
        """
        Provides the Hash to use.
        """

        return hashes.Hash(getattr(hashes, self.algo)(), backend=default_backend())

    def hash_file(self, file_path: str) -> str:
        """
        Hashes the content of the given file.

        :param file_path:
            The path of the file to read.
        """

        block_size = 4096

        digest = self.__get_hash()

        with FileHelper(file_path).open("rb") as file_stream:
            block = file_stream.read(block_size)

            while block:
                digest.update(block)
                block = file_stream.read(block_size)

        return digest.finalize().hex()

    def hash_data(self, data: Union[str, bytes]) -> str:
        """
        Hashes the given data.

        :param data:
            The data to hash.

        :raise TypeError:
            When :code:`data` is not :py:class:`str` or :py:class:`bytes`.
        """

        if not isinstance(data, (bytes, str)):
            raise TypeError(f"<data> should be {bytes} or {str}, {type(data)}, given.")

        if isinstance(data, str):
            data = data.encode()

        digest = self.__get_hash()
        digest.update(data)

        return digest.finalize().hex()



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the base of all mariadb related migrations.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import functools
from typing import Any, Generator, Tuple

import PyFunceble.cli.facility
import PyFunceble.cli.factory
import PyFunceble.sessions
from PyFunceble.cli.migrators.base import MigratorBase


class MariaDBMigratorBase(MigratorBase):
    """
    Provides the base of all our mariadb migration.
    """

    def execute_if_authorized(default: Any = None):  # pylint: disable=no-self-argument
        """
        Executes the decorated method only if we are authorized to process.
        Otherwise, apply the given :code:`default`.
        """

        def inner_metdhod(func):
            @functools.wraps(func)
            def wrapper(self, *args, **kwargs):
                if self.authorized:
                    return func(self, *args, **kwargs)  # pylint: disable=not-callable
                return self if default is None else default

            return wrapper

        return inner_metdhod

    def does_table_exists(self, name: str) -> bool:
        """
        Checks if the given table name exists.
        """

        statement = (
            "SELECT COUNT(*) "
            "FROM information_schema.tables "
            "WHERE table_schema = :database_name "
            "AND table_name = :table_name "
        )

        result = self.db_session.execute(
            statement,
            {
                "database_name": PyFunceble.cli.factory.DBSession.credential.name,
                "table_name": name,
            },
        ).fetchone()

        return result["COUNT(*)"] == 1

    def get_rows(
        self, statement: str, limit: int = 20
    ) -> Generator[Tuple[str, int], dict, None]:
        """
        Run the given statement with a defined limit, and yield each row.

        .. warning::
            If you don't delete the given rows, this method will be infinite.
        """

        statement += f" LIMIT {limit}"

        while True:
            db_result = list(self.db_session.execute(statement).fetchall())

            if not db_result:
                break

            for result in db_result:
                yield dict(result)

    @property
    def authorized(self) -> bool:
        """
        Provides the authorization to process.
        """

        return PyFunceble.cli.facility.CredentialLoader.is_already_loaded()

    @execute_if_authorized(None)
    def migrate(self) -> "MariaDBMigratorBase":
        """
        Provides the migration (itself).
        """

        raise NotImplementedError()

    @execute_if_authorized(None)
    def start(self) -> "MariaDBMigratorBase":
        """
        Starts the migration if wanted.
        """

        PyFunceble.facility.Logger.info("Started migration.")

        self.migrate()

        PyFunceble.facility.Logger.info("Finished migration.")

        return self



"""Make id a bigint.

Revision ID: bef7bcaac3f2
Revises: 3a4c55a9320d
Create Date: 2020-12-16 19:09:41.212679

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import mysql

# pylint: skip-file

# revision identifiers, used by Alembic.
revision = "bef7bcaac3f2"
down_revision = "3a4c55a9320d"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column(
        "pyfunceble_continue",
        "id",
        existing_type=mysql.INTEGER(display_width=11),
        type_=sa.BigInteger(),
        autoincrement=True,
    )
    op.alter_column(
        "pyfunceble_inactive",
        "id",
        existing_type=mysql.INTEGER(display_width=11),
        type_=sa.BigInteger(),
        autoincrement=True,
    )
    op.alter_column(
        "pyfunceble_whois_record",
        "id",
        existing_type=mysql.INTEGER(display_width=11),
        type_=sa.BigInteger(),
        autoincrement=True,
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column(
        "pyfunceble_whois_record",
        "id",
        existing_type=sa.BigInteger(),
        type_=mysql.INTEGER(display_width=11),
        autoincrement=True,
    )
    op.alter_column(
        "pyfunceble_inactive",
        "id",
        existing_type=sa.BigInteger(),
        type_=mysql.INTEGER(display_width=11),
        autoincrement=True,
    )
    op.alter_column(
        "pyfunceble_continue",
        "id",
        existing_type=sa.BigInteger(),
        type_=mysql.INTEGER(display_width=11),
        autoincrement=True,
    )
    # ### end Alembic commands ###



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the base of all our CI classes.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# pylint: disable=too-many-lines

import datetime
import functools
import secrets
from typing import Any, Optional

import PyFunceble.cli.continuous_integration.exceptions
import PyFunceble.facility
import PyFunceble.storage
from PyFunceble.helpers.command import CommandHelper
from PyFunceble.helpers.environment_variable import EnvironmentVariableHelper
from PyFunceble.helpers.regex import RegexHelper


class ContinuousIntegrationBase:
    """
    Provides the base of all continuous integration methods.

    :param authorized:
        The authorization to run.
    :param git_email:
        The email to apply while initilizing the git repository for push.
    :param git_name:
        The name to apply while initilizing the git repository for push.
    :param git_branch:
        The branch to use while testing.
    :param git_distribution_branch:
        The branch to push the results into.
    :param token:
        The token to apply while initilizing the git repository for push.
    :param command:
        The command to execute before each push (except the latest one).
    :param end_command:
        The commant to execute at the very end.
    :param commit_message:
        The commit message to apply before each push (except the latest one).
    :param end_commit_message:
        The commit message to apply at the very end.
    :param max_exec_minutes:
        The maximum of minutes to apply before considering the current
        session as finished.
    """

    STD_AUTHORIZED: bool = False
    STD_GIT_EMAIL: Optional[str] = None
    STD_GIT_NAME: Optional[str] = None
    STD_GIT_BRANCH: Optional[str] = "master"
    STD_GIT_DISTRIBUTION_BRANCH: Optional[str] = "master"
    STD_COMMAND: Optional[str] = None
    STD_END_COMMAND: Optional[str] = None
    STD_COMMIT_MESSAGE: str = "PyFunceble - AutoSave"
    STD_END_COMMIT_MESSAGE: str = "PyFunceble - Results"
    STD_MAX_EXEC_MINUTES: int = 15

    end_commit_marker: str = "[ci skip]"

    _authorized: bool = False
    _git_email: Optional[str] = None
    _git_name: Optional[str] = None
    _git_branch: Optional[str] = None
    _git_distribution_branch: Optional[str] = None
    _token: Optional[str] = None
    _command: Optional[str] = None
    _end_command: Optional[str] = None
    _commit_message: Optional[str] = None
    _end_commit_message: Optional[str] = None
    _max_exec_minutes: Optional[int] = None

    start_time: Optional[datetime.datetime] = None
    expected_end_time: Optional[datetime.datetime] = None

    git_initialized: bool = False

    def __init__(
        self,
        *,
        authorized: Optional[bool] = None,
        git_email: Optional[str] = None,
        git_name: Optional[str] = None,
        git_branch: Optional[str] = None,
        git_distribution_branch: Optional[str] = None,
        token: Optional[str] = None,
        command: Optional[str] = None,
        end_command: Optional[str] = None,
        commit_message: Optional[str] = None,
        end_commit_message: Optional[str] = None,
        max_exec_minutes: Optional[int] = None,
    ) -> None:
        if authorized is not None:
            self.authorized = authorized
        else:
            self.guess_and_set_authorized()

        if git_email is not None:
            self.git_email = git_email
        else:
            self.guess_and_set_git_email()

        if git_name is not None:
            self.git_name = git_name
        else:
            self.guess_and_set_git_name()

        if git_branch is not None:
            self.git_branch = git_branch
        else:
            self.guess_and_set_git_branch()

        if git_distribution_branch is not None:
            self.git_distribution_branch = git_distribution_branch
        else:
            self.guess_and_set_git_distribution_branch()

        if token is not None:
            self.token = token
        else:
            self.guess_and_set_token()

        if command is not None:
            self.command = command
        else:
            self.guess_and_set_command()

        if end_command is not None:
            self.end_command = end_command
        else:
            self.guess_and_set_end_command()

        if commit_message is not None:
            self.commit_message = commit_message
        else:
            self.guess_and_set_commit_message()

        if end_commit_message is not None:
            self.end_commit_message = end_commit_message
        else:
            self.guess_and_set_end_commit_message()

        if max_exec_minutes is not None:
            self.max_exec_minutes = max_exec_minutes
        else:
            self.guess_and_set_max_exec_minutes()

    def execute_if_authorized(default: Any = None):  # pylint: disable=no-self-argument
        """
        Executes the decorated method only if we are authorized to process.
        Otherwise, apply the given :code:`default`.
        """

        def inner_metdhod(func):
            @functools.wraps(func)
            def wrapper(self, *args, **kwargs):
                if self.authorized:
                    return func(self, *args, **kwargs)  # pylint: disable=not-callable
                return self if default is None else default

            return wrapper

        return inner_metdhod

    def ensure_git_email_is_given(func):  # pylint: disable=no-self-argument
        """
        Ensures that the Git Email is given before launching the decorated
        method.

        :raise PyFunceble.cli.continuous_integration.exceptions.GitEmailNotFound:
            When the Git Email is not found.
        """

        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            if not self.git_email:
                # pylint: disable=line-too-long
                raise PyFunceble.cli.continuous_integration.exceptions.GitEmailNotFound()

            return func(self, *args, **kwargs)  # pylint: disable=not-callable

        return wrapper

    def ensure_git_name_is_given(func):  # pylint: disable=no-self-argument
        """
        Ensures that the Git Name is given before launching the decorated
        method.

        :raise PyFunceble.cli.continuous_integration.exceptions.GitNameNotFound:
            When the Git Name is not found.
        """

        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            if not self.git_name:
                raise PyFunceble.cli.continuous_integration.exceptions.GitNameNotFound()

            return func(self, *args, **kwargs)  # pylint: disable=not-callable

        return wrapper

    def ensure_git_branch_is_given(func):  # pylint: disable=no-self-argument
        """
        Ensures that the Git Branch is given before launching the decorated
        method.

        :raise PyFunceble.cli.continuous_integration.exceptions.GitBranchNotFound:
            When the Git Branch is not found.
        """

        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            if not self.git_name:
                raise PyFunceble.cli.continuous_integration.exceptions.GitNameNotFound()

            return func(self, *args, **kwargs)  # pylint: disable=not-callable

        return wrapper

    def ensure_git_distribution_branch_is_given(
        func,
    ):  # pylint: disable=no-self-argument,line-too-long
        """
        Ensures that the Git distribution Branch is given before launching the
        decorated method.

        :raise PyFunceble.cli.continuous_integration.exceptions.GitDistributionBranchNotFound:
            When the Git distribution Branch is not found.
        """

        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            if not self.git_name:
                raise PyFunceble.cli.continuous_integration.exceptions.GitNameNotFound()

            return func(self, *args, **kwargs)  # pylint: disable=not-callable

        return wrapper

    def ensure_token_is_given(func):  # pylint: disable=no-self-argument
        """
        Ensures that the token is given before launching the decorated method.

        :raise PyFunceble.cli.continuous_integration.exceptions.TokenNotFound:
            When the token is not found.
        """

        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            if not self.token:
                raise PyFunceble.cli.continuous_integration.exceptions.TokenNotFound()

            return func(self, *args, **kwargs)  # pylint: disable=not-callable

        return wrapper

    def ensure_start_time_is_given(func):  # pylint: disable=no-self-argument
        """
        Ensures that the starting time is given before launching the decorated
        method.

        :raise PyFunceble.cli.continuous_integration.exceptions.StartTimeNotFound:
            When the token is not found.
        """

        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            if not self.start_time:
                # pylint: disable=line-too-long
                raise PyFunceble.cli.continuous_integration.exceptions.StartTimeNotFound()

            return func(self, *args, **kwargs)  # pylint: disable=not-callable

        return wrapper

    @property
    def authorized(self) -> Optional[bool]:
        """
        Provides the currently state of the :code:`_authorized` attribute.
        """

        return self._authorized

    @authorized.setter
    def authorized(self, value: bool) -> None:
        """
        Sets the value of the :code:`authorized` attribute.

        :param value:
            The value to set.

        :raise TypeError:
            When the given :code:`value` is not a :py:class:`bool`
        """

        if not isinstance(value, bool):
            raise TypeError(f"<value> should be {bool}, {type(value)} given.")

        self._authorized = value

    def set_authorized(self, value: bool) -> "ContinuousIntegrationBase":
        """
        Sets the value of the :code:`authorized` attribute.

        :param value:
            The value to set.
        """

        self.authorized = value

        return self

    @property
    def git_email(self) -> Optional[str]:
        """
        Provides the currently state of the :code:`_git_email` attribute.
        """

        return self._git_email

    @git_email.setter
    def git_email(self, value: str) -> None:
        """
        Sets the Git Email to use.

        :param value:
            The value to set.

        :raise TypeError:
            When the given :code:`value` is not a :py:class:`str`.
        :raise valueError:
            When the given :code:`value` is empty.
        """

        if not isinstance(value, str):
            raise TypeError(f"<value> should be {str}, {type(value)} given.")

        if not value:
            raise ValueError("<value> should not be empty.")

        self._git_email = value

    def set_git_email(self, value: str) -> "ContinuousIntegrationBase":
        """
        Sets the Git Email to use.

        :param value:
            The value to set.
        """

        self.git_email = value

        return self

    @property
    def git_name(self) -> Optional[str]:
        """
        Provides the current state of the :code:`_git_name` attribute.
        """

        return self._git_name

    @git_name.setter
    def git_name(self, value: str) -> None:
        """
        Sets the Git Name to use.

        :param value:
            The value to set.

        :raise TypeError:
            When the given :code:`value` is not a :py:class:`str`.
        :raise valueError:
            When the given :code:`value` is empty.
        """

        if not isinstance(value, str):
            raise TypeError(f"<value> should be {str}, {type(value)} given.")

        if not value:
            raise ValueError("<value> should not be empty.")

        self._git_name = value

    def set_git_name(self, value: str) -> "ContinuousIntegrationBase":
        """
        Sets the Git Name to use.

        :param value:
            The value to set.
        """

        self.git_name = value

        return self

    @property
    def git_branch(self) -> Optional[str]:
        """
        Provides the current state of the :code:`_git_branch` attribute.
        """

        return self._git_branch

    @git_branch.setter
    def git_branch(self, value: str) -> None:
        """
        Sets the Git Branch to use.

        :param value:
            The value to set.

        :raise TypeError:
            When the given :code:`value` is not a :py:class:`str`.
        :raise valueError:
            When the given :code:`value` is empty.
        """

        if not isinstance(value, str):
            raise TypeError(f"<value> should be {str}, {type(value)} given.")

        if not value:
            raise ValueError("<value> should not be empty.")

        self._git_branch = value

    def set_git_branch(self, value: str) -> "ContinuousIntegrationBase":
        """
        Sets the Git Branch to use.

        :param value:
            The value to set.
        """

        self.git_branch = value

        return self

    @property
    def git_distribution_branch(self) -> Optional[str]:
        """
        Provides the current state of the :code:`_git_distribution_branch`
        attribute.
        """

        return self._git_distribution_branch

    @git_distribution_branch.setter
    def git_distribution_branch(self, value: str) -> None:
        """
        Sets the Git distribution Branch to use.

        :param value:
            The value to set.

        :raise TypeError:
            When the given :code:`value` is not a :py:class:`str`.
        :raise valueError:
            When the given :code:`value` is empty.
        """

        if not isinstance(value, str):
            raise TypeError(f"<value> should be {str}, {type(value)} given.")

        if not value:
            raise ValueError("<value> should not be empty.")

        self._git_distribution_branch = value

    def set_git_distribution_branch(self, value: str) -> "ContinuousIntegrationBase":
        """
        Sets the Git distribution Branch to use.

        :param value:
            The value to set.
        """

        self.git_distribution_branch = value

        return self

    @property
    def token(self) -> Optional[str]:
        """
        Provides the current state of the :code:`_token` attribute.
        """

        return self._token

    @token.setter
    def token(self, value: str) -> None:
        """
        Sets the token to use.

        :param value:
            The value to set.

        :raise TypeError:
            When the given :code:`value` is not a :py:class:`str`.
        :raise valueError:
            When the given :code:`value` is empty.
        """

        if not isinstance(value, str):
            raise TypeError(f"<value> should be {str}, {type(value)} given.")

        if not value:
            raise ValueError("<value> should not be empty.")

        self._token = value

    def set_token(self, value: str) -> "ContinuousIntegrationBase":
        """
        Sets the token to use.

        :param value:
            The value to set.
        """

        self.token = value

        return value

    @property
    def command(self) -> Optional[str]:
        """
        Provides the current state of the :code:`_command` attribute.
        """

        return self._command

    @command.setter
    def command(self, value: str) -> None:
        """
        Sets the command to work with.

        :param value:
            The command to set.

        :raise TypeError:
            When the given :code:`value` is not a :py:class:`str`.
        :raise ValueError:
            When the given :code:`value` is empty.
        """

        if not isinstance(value, str):
            raise TypeError(f"<value> should be {str}, {type(value)} given.")

        if not value:
            raise ValueError("<value> should not be empy.")

        self._command = value

    def set_command(self, value: str) -> "ContinuousIntegrationBase":
        """
        Sets the command to work with.

        :param value:
            The command to set.
        """

        self.command = value

        return self

    @property
    def end_command(self) -> Optional[str]:
        """
        Provides the current state of the :code:`_end_command` attribute.
        """

        return self._end_command

    @end_command.setter
    def end_command(self, value: str) -> None:
        """
        Sets the command to execute at the really end of the process with.

        :param value:
            The command to set.

        :raise TypeError:
            When the given :code:`value` is not a :py:class:`str`.
        :raise ValueError:
            When the given :code:`value` is empty.
        """

        if not isinstance(value, str):
            raise TypeError(f"<value> should be {str}, {type(value)} given.")

        if not value:
            raise ValueError("<value> should not be empy.")

        self._end_command = value

    def set_end_command(self, value: str) -> "ContinuousIntegrationBase":
        """
        Sets the command to execute at the really end of the process with.

        :param value:
            The command to set.
        """

        self.end_command = value

        return self

    @property
    def commit_message(self) -> Optional[str]:
        """
        Provides the current state of the :code:`_commit_message` attribute.
        """

        return self._commit_message

    @commit_message.setter
    def commit_message(self, value: str) -> None:
        """
        Sets the commit message to apply to all commits except the final one.

        :param value:
            The message to set.

        :raise TypeError:
            When the given :code:`value` is not a :py:class:`str`.
        :raise ValueError:
            When the given :code:`value` is empty.
        """

        if not isinstance(value, str):
            raise TypeError(f"<value> should be {str}, {type(value)} given.")

        if not value:
            raise ValueError("<value> should not be empy.")

        self._commit_message = value

    def set_commit_message(self, value: str) -> "ContinuousIntegrationBase":
        """
        Sets the commit message to apply to all commits except the final one.

        :param value:
            The message to set.
        """

        self.commit_message = value

        return self

    @property
    def end_commit_message(self) -> Optional[str]:
        """
        Provides the current state of the :code:`_end_commit_message` attribute.
        """

        return self._end_commit_message

    @end_commit_message.setter
    def end_commit_message(self, value: str) -> None:
        """
        Sets the commit message to apply to the final one.

        :param value:
            The message to set.

        :raise TypeError:
            When the given :code:`value` is not a :py:class:`str`.
        :raise ValueError:
            When the given :code:`value` is empty.
        """

        if not isinstance(value, str):
            raise TypeError(f"<value> should be {str}, {type(value)} given.")

        if not value:
            raise ValueError("<value> should not be empy.")

        self._end_commit_message = value

    def set_end_commit_message(self, value: str) -> "ContinuousIntegrationBase":
        """
        Sets the commit message to apply to the final one.

        :param value:
            The command to set.
        """

        self.end_commit_message = value

        return self

    @property
    def max_exec_minutes(self) -> Optional[str]:
        """
        Provides the current state of the :code:`_max_exec_minutes` attribute.
        """

        return self._max_exec_minutes

    @max_exec_minutes.setter
    def max_exec_minutes(self, value: str) -> None:
        """
        Sets the maximum waiting time before considering the time as exceeded.

        :param value:
            The message to set.

        :raise TypeError:
            When the given :code:`value` is not a :py:class:`int`.
        :raise ValueError:
            When the given :code:`value` is less than :code:`1`.
        """

        if not isinstance(value, int):
            raise TypeError(f"<value> should be {int}, {type(value)} given.")

        if value < 1:
            raise ValueError("<value> should be greater or equal to 1.")

        self._max_exec_minutes = value

    def set_max_exec_minutes(self, value: str) -> "ContinuousIntegrationBase":
        """
        Sets the maximum waiting time before considering the time as exceeded.

        :param value:
            The command to set.
        """

        self._max_exec_minutes = value

        return self

    @execute_if_authorized(None)
    def set_start_time(self) -> "ContinuousIntegrationBase":
        """
        Sets the starting time to now.
        """

        self.start_time = datetime.datetime.utcnow()
        self.expected_end_time = self.start_time + datetime.timedelta(
            minutes=self.max_exec_minutes
        )

    @staticmethod
    def get_remote_destination():
        """
        Provides the remote destination to use.

        :raise PyFunceble.cli.continuous_integration.exceptions.RemoteURLNotFound:
            When we could not determine the remote destination.
        """

        regex = r"(?:[a-z]+(?:\s+|\t+))(.*)(?:(?:\s+|\t+)\([a-z]+\))"
        remote_of_interest = [
            x
            for x in CommandHelper("git remote -v").execute().splitlines()
            if "(fetch)" in x
        ][0]

        filtered = RegexHelper(regex).match(
            remote_of_interest, return_match=True, group=1
        )

        if filtered:
            if "@" in filtered:
                return filtered[filtered.find("@") + 1 :]
            if "//" in filtered:
                return filtered[filtered.find("//") + 2 :]

        raise PyFunceble.cli.continuous_integration.exceptions.RemoteURLNotFound()

    @staticmethod
    def exec_command(command: str, allow_stdout: bool) -> None:
        """
        Exceutes the given command.

        :param command:
            The command to execute.

        :param allow_stdout:
            Allows us to return the command output to stdout.
        """

        PyFunceble.facility.Logger.debug("Executing %r", command)

        command_helper = CommandHelper(command)

        if allow_stdout:
            command_helper.run_to_stdout()
        else:
            command_helper.execute()

    def guess_and_set_authorized(self) -> "ContinuousIntegrationBase":
        """
        Tries to guess the authorization.
        """

        if PyFunceble.facility.ConfigLoader.is_already_loaded():
            self.authorized = bool(
                PyFunceble.storage.CONFIGURATION.cli_testing.ci.active
            )
        else:
            self.authorized = self.STD_AUTHORIZED

        return self

    def guess_and_set_git_email(self) -> "ContinuousIntegrationBase":
        """
        Tries to guess and set the Git Email.
        """

        environment_var = EnvironmentVariableHelper("GIT_EMAIL")

        if environment_var.exists():
            self.git_email = environment_var.get_value()

        return self

    def guess_and_set_git_name(self) -> "ContinuousIntegrationBase":
        """
        Tries to guess and set the Git Name.
        """

        environment_var = EnvironmentVariableHelper("GIT_NAME")

        if environment_var.exists():
            self.git_name = environment_var.get_value()

        return self

    def guess_and_set_git_branch(self) -> "ContinuousIntegrationBase":
        """
        Tries to guess and set the Git Branch.
        """

        environment_var = EnvironmentVariableHelper("GIT_BRANCH")

        if environment_var.exists():
            self.git_branch = environment_var.get_value()
        elif PyFunceble.facility.ConfigLoader.is_already_loaded():
            self.git_branch = PyFunceble.storage.CONFIGURATION.cli_testing.ci.branch
        else:
            self.git_branch = self.STD_GIT_BRANCH

        return self

    def guess_and_set_git_distribution_branch(self) -> "ContinuousIntegrationBase":
        """
        Tries to guess and set the Git distribution Branch.
        """

        environment_var = EnvironmentVariableHelper("GIT_DISTRIBUTION_BRANCH")

        if environment_var.exists():
            self.git_distribution_branch = environment_var.get_value()
        elif PyFunceble.facility.ConfigLoader.is_already_loaded():
            self.git_distribution_branch = (
                PyFunceble.storage.CONFIGURATION.cli_testing.ci.distribution_branch
            )
        else:
            self.git_distribution_branch = self.STD_GIT_DISTRIBUTION_BRANCH

        return self

    def guess_and_set_token(self) -> "ContinuousIntegrationBase":
        """
        Tries to guess and set the token.
        """

        raise NotImplementedError()

    def guess_and_set_command(self) -> "ContinuousIntegrationBase":
        """
        Tries to guess and set the command to execute.
        """

        if PyFunceble.facility.ConfigLoader.is_already_loaded():
            if PyFunceble.storage.CONFIGURATION.cli_testing.ci.command:
                self.command = PyFunceble.storage.CONFIGURATION.cli_testing.ci.command

        return self

    def guess_and_set_end_command(self) -> "ContinuousIntegrationBase":
        """
        Tries to guess and set the command to execute at the very end.
        """

        if PyFunceble.facility.ConfigLoader.is_already_loaded():
            if PyFunceble.storage.CONFIGURATION.cli_testing.ci.end_command:
                self.end_command = (
                    PyFunceble.storage.CONFIGURATION.cli_testing.ci.end_command
                )

        return self

    def guess_and_set_commit_message(self) -> "ContinuousIntegrationBase":
        """
        Tries to guess and set the commit message to apply.
        """

        if PyFunceble.facility.ConfigLoader.is_already_loaded():
            if PyFunceble.storage.CONFIGURATION.cli_testing.ci.commit_message:
                self.commit_message = (
                    PyFunceble.storage.CONFIGURATION.cli_testing.ci.commit_message
                )
            else:
                self.commit_message = self.STD_COMMIT_MESSAGE
        else:
            self.commit_message = self.STD_COMMIT_MESSAGE

        return self

    def guess_and_set_end_commit_message(self) -> "ContinuousIntegrationBase":
        """
        Tries to guess and set the commit message to apply at the very end.
        """

        if PyFunceble.facility.ConfigLoader.is_already_loaded():
            if PyFunceble.storage.CONFIGURATION.cli_testing.ci.end_commit_message:
                self.end_commit_message = (
                    PyFunceble.storage.CONFIGURATION.cli_testing.ci.end_commit_message
                )
            else:
                self.end_commit_message = self.STD_END_COMMIT_MESSAGE
        else:
            self.end_commit_message = self.STD_END_COMMIT_MESSAGE

        return self

    def guess_and_set_max_exec_minutes(self) -> "ContinuousIntegrationBase":
        """
        Tries to guess and set the maximum execution time.
        """

        if PyFunceble.facility.ConfigLoader.is_already_loaded():
            self.max_exec_minutes = (
                PyFunceble.storage.CONFIGURATION.cli_testing.ci.max_exec_minutes
            )
        else:
            self.max_exec_minutes = self.STD_MAX_EXEC_MINUTES

        return self

    def guess_all_settings(self) -> "ContinuousIntegrationBase":
        """
        Try to guess all settings.
        """

        to_ignore = ["guess_all_settings"]

        for method in dir(self):
            if method in to_ignore or not method.startswith("guess_"):
                continue

            getattr(self, method)()

        return self

    def is_authorized(self) -> bool:
        """
        Checks if the current object is authorized to run.
        """

        return self.authorized is True

    @execute_if_authorized(False)
    @ensure_start_time_is_given
    def is_time_exceeded(self) -> bool:
        """
        Checks if we exceeded the allocated time we have.
        """

        return self.expected_end_time < datetime.datetime.utcnow()

    @execute_if_authorized(None)
    @ensure_token_is_given
    def init_git_remote_with_token(self) -> "ContinuousIntegrationBase":
        """
        Initiates the git remote URL with the help of the given token.
        """

        remote = self.get_remote_destination()

        # Each entries should have a tuple containing the command to run and
        # if we are authorized to print to STDOUT.
        commands_to_execute = [
            ("git remote rm origin", False),
            ("git remote add origin " f"https://{self.token}@{remote}", False),
            ("git remote update", False),
            ("git fetch", False),
        ]

        for command, allow_stdout in commands_to_execute:
            self.exec_command(command, allow_stdout)

        return self

    @execute_if_authorized(None)
    @ensure_git_name_is_given
    @ensure_git_email_is_given
    @ensure_git_branch_is_given
    @ensure_git_distribution_branch_is_given
    def init_git(self) -> "ContinuousIntegrationBase":
        """
        Initiate the git repository.
        """

        PyFunceble.facility.Logger.info("Started initialization of GIT.")

        if self.token:
            self.init_git_remote_with_token()

        commands = [
            (f'git config --local user.email "{self.git_email}"', False),
            (f'git config --local user.name "{self.git_name}"', False),
            ("git config --local push.default simple", False),
            ("git config --local pull.rebase true", False),
            ("git config --local core.autocrlf true", False),
            ("git config --local branch.autosetuprebase always", False),
            (f'git checkout "{self.git_branch}"', False),
            ("git fetch origin", False),
            (
                f"git rebase -X theirs " f"origin/{self.git_distribution_branch}",
                False,
            ),
        ]

        for command, allow_stdout in commands:
            self.exec_command(command, allow_stdout)

        self.git_initialized = True

        PyFunceble.facility.Logger.info("Finished initialization of GIT.")

        return self

    def fix_permissions(self) -> "ContinuousIntegrationBase":
        """
        A method to overwrite when custom rules for permissions are needed.

        .. note::
            This method is automatically called by the methods who apply
            commits.
        """

        return self

    @execute_if_authorized(None)
    def push_changes(self, branch: str, *, exit_it: bool = True) -> None:
        """
        Pushes the changes.

        :param branch:
            The branch to push.

        :param exit_it:
            Exits after the push ?

        :raise PyFunceble.cli.continuous_integration.exceptions.StopExecution:
            When the :code:`exit_it` is set to :code:`True`.
        """

        PyFunceble.facility.Logger.info("Started to push GIT changes.")

        commands = []

        if self.git_initialized:
            commands.extend(
                [
                    (
                        f'git checkout "{branch}"',
                        True,
                    ),
                    ("git fetch origin", True),
                    (f"git rebase -X theirs origin/{branch}", True),
                    (f"git push origin {branch}", True),
                ]
            )

        for command, allow_stdout in commands:
            self.exec_command(command, allow_stdout)

        PyFunceble.facility.Logger.info("Finished to push GIT changes.")

        if exit_it:
            raise PyFunceble.cli.continuous_integration.exceptions.StopExecution()

    @execute_if_authorized(None)
    def apply_end_commit(self) -> None:
        """
        Apply the "end" commit and push.

        Side effect:
            It runs the declared command to execute.
        """

        PyFunceble.facility.Logger.info(
            "Started to prepare and apply final GIT commit."
        )

        self.fix_permissions()

        commands = []

        if self.git_distribution_branch != self.git_branch:
            branch_to_use = self.git_distribution_branch
        else:
            branch_to_use = self.git_branch

        if self.end_command:
            commands.append((self.end_command, True))

        if self.git_initialized:
            commands.extend(
                [
                    ("git fetch origin", True),
                    (
                        f"git rebase -X theirs origin/{branch_to_use}",
                        True,
                    ),
                    ("git add --all", True),
                    (
                        "git commit -a -m "
                        f'"{self.end_commit_message} {self.end_commit_marker}\n\n'
                        f'{secrets.token_urlsafe(18)}"',
                        True,
                    ),
                ]
            )

        for command, allow_stdout in commands:
            self.exec_command(command, allow_stdout)

        if self.end_command:
            # Fix permissions because we met some strange behaviors in the past.
            self.fix_permissions()

        PyFunceble.facility.Logger.info(
            "Finished to prepare and apply final GIT commit."
        )

        if self.git_distribution_branch != self.git_branch:
            self.push_changes(branch_to_use)
        else:
            self.push_changes(branch_to_use)

    @execute_if_authorized(None)
    def apply_commit(self) -> None:
        """
        Apply the commit and push.

        Side effect:
            It runs the declared command to execute.
        """

        PyFunceble.facility.Logger.info("Started to prepare and apply GIT commit.")

        self.fix_permissions()

        commands = []

        if self.command:
            commands.append((self.command, True))

        if self.git_initialized:
            commands.extend(
                [
                    ("git add --all", True),
                    (
                        "git commit -a -m "
                        f'"{self.commit_message}\n\n'
                        f'{secrets.token_urlsafe(18)}"',
                        True,
                    ),
                ]
            )

        for command, allow_stdout in commands:
            self.exec_command(command, allow_stdout)

        if self.command:
            # Fix permissions because we met some strange behaviors in the past.
            self.fix_permissions()

        PyFunceble.facility.Logger.info("Finished to prepare and apply GIT commit.")

        self.push_changes(self.git_branch)

    @execute_if_authorized(None)
    def bypass(self) -> None:
        """
        Stops everything if the latest commit message match any of those:

            - :code:`[PyFunceble skip]` (case insensitive)
            - :code:`[PyFunceble-skip]` (case insensitive)
            - :attr:`~PyFunceble.cli.continuous_integration.base.end_commit_marker`
        """

        our_marker = ["[pyfunceble skip]", "[pyfunceble-skip]", self.end_commit_marker]
        latest_commit = CommandHelper("git log -1").execute().lower()

        if any(x.lower() in latest_commit for x in our_marker):
            PyFunceble.facility.Logger.info(
                "Bypass marker caught. Saving and stopping process."
            )

            raise PyFunceble.cli.continuous_integration.exceptions.StopExecution()

    @execute_if_authorized(None)
    def init(self) -> "ContinuousIntegrationBase":
        """
        Initiate our infrastructure for the current CI engine.

        The purpose of this method is to be able to have some custom init based
        on the CI we are currently on.

        The init method should be manually started before runing any further
        action.

        .. warning::
            We assume that we are aware that you should run this method first.
        """

        PyFunceble.facility.Logger.info("Started initizalization of workflow.")

        self.init_git()

        PyFunceble.facility.Logger.info("Finished initizalization of workflow.")

        return self



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides our credential loader.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import functools
from getpass import getpass
from typing import Any, Optional

import PyFunceble.cli.factory
import PyFunceble.facility
import PyFunceble.storage
from PyFunceble.database.credential.base import CredentialBase
from PyFunceble.database.credential.mariadb import MariaDBCredential
from PyFunceble.database.credential.mysql import MySQLCredential
from PyFunceble.helpers.environment_variable import EnvironmentVariableHelper


class CredentialLoader:
    """
    Provides our creadential loader.

    The idea is to have an interface which provides the credential to use
    but at the same time load and initiate the credential interface after
    loading it from the user input or the environment variables.

    :param db_type:
        The database type to load the credential for.
    """

    DB_TYPE2OBJ: dict = {
        "csv": None,
        "mysql": MySQLCredential,
        "mariadb": MariaDBCredential,
    }

    credential: Optional[CredentialBase] = None
    """
    The credential (itself).
    """

    _db_type: Optional[str] = None

    def __init__(self, db_type: Optional[str] = None) -> None:
        if db_type is not None:
            self.db_type = db_type
        elif PyFunceble.facility.ConfigLoader.is_already_loaded():
            self.db_type = PyFunceble.storage.CONFIGURATION.cli_testing.db_type

    def ensure_db_type_is_given(func):  # pylint: disable=no-self-argument
        """
        Ensure that the db type is given before launching the decorated
        method.

        :raise TypeError:
            When the db type is not given.
        """

        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            if not isinstance(self.db_type, str):
                if PyFunceble.facility.ConfigLoader.is_already_loaded():
                    self.db_type = PyFunceble.storage.CONFIGURATION.cli_testing.db_type
                else:
                    raise TypeError(f"<self.db_type> should be {str}.")

            return func(self, *args, **kwargs)  # pylint: disable=not-callable

        return wrapper

    def execute_if_authorized(default: Any = None):  # pylint: disable=no-self-argument
        """
        Executes the decorated method only if we are authorized to process.
        Otherwise, apply the given :code:`default`.
        """

        def inner_metdhod(func):
            @functools.wraps(func)
            def wrapper(self, *args, **kwargs):
                if self.authorized:
                    return func(self, *args, **kwargs)  # pylint: disable=not-callable
                return self if default is None else default

            return wrapper

        return inner_metdhod

    def start_if_not_started(func):  # pylint: disable=no-self-argument
        """
        Launches the start method before launching the decorated method.
        """

        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            if not isinstance(self.credential, CredentialBase):
                self.start()

            return func(self, *args, **kwargs)  # pylint: disable=not-callable

        return wrapper

    @property
    def authorized(self) -> bool:
        """
        Provides the authorization to run.
        """

        return self.db_type != "csv"

    @property
    def db_type(self) -> Optional[str]:
        """
        Provides the current state of the :code:`_db_type` attribute.
        """

        return self._db_type

    @db_type.setter
    def db_type(self, value: str) -> None:
        """
        Sets the database type to work with.

        :param value:
            The value to set.

        :raise TypeError:
            When the given :code:`value` is not a :py:class:`str`.
        :raise ValueError:
            When the given :code:`value` is empty or not supported.
        """

        if not isinstance(value, str):
            raise TypeError(f"<value> should be {str}, {type(value)} given.")

        if not value:
            raise ValueError("<value> should not be empty.")

        value = value.lower()

        if value not in self.DB_TYPE2OBJ:
            raise ValueError(f"<value> ({value!r}) is not supported.")

        self._db_type = value

    def set_db_type(self, value: str) -> "CredentialLoader":
        """
        Sets the database type to work with.

        :param value:
            The value to set.
        """

        self.db_type = value

        return self

    @execute_if_authorized(False)
    def is_already_loaded(self) -> bool:
        """
        Checks if the credential was already loaded.
        """

        return isinstance(self.credential, CredentialBase)

    def set_credential_var(self, var_name: str, value: Any) -> "CredentialLoader":
        """
        Sets the given :code:`var_name` of the credential object with the given
        :code:`value`.

        :param var_name:
            The name of the variable to set.
        :param value:
            The value of to set.
        """

        try:
            setattr(self.credential, var_name, value)
        except TypeError:
            setattr(self.credential, var_name, int(value))

    def ask_for_info(self, var_name: str, default: Any) -> str:
        """
        Asks the credential to the end-user.

        :param var_name:
            The name of the part to ask for example :code:`host` or
            :code:`password`.
        :param default:
            The default value to return if the user don't give us anything.
        """

        message = (
            f"[{self.db_type.upper()}] Please give us your DB {var_name} "
            f"(Default: {default!r}): "
        )

        if var_name == "password":
            user_input = getpass(message)
        else:
            user_input = input(message)

        if user_input:
            return user_input
        return default

    @execute_if_authorized(None)
    @ensure_db_type_is_given
    def start(self, *, ignore_cli: bool = True) -> "CredentialLoader":
        """
        Starts the loading of the credential.

        :param ignore_cli:
            Ignore questions to end-user.
        """

        if not isinstance(self.credential, CredentialBase) and self.authorized:

            # We directly share the credential object into the DBSession object.
            # This will let us use the DBSession without having to think about
            # any other headache.
            self.credential = (
                PyFunceble.cli.factory.DBSession.credential
            ) = self.DB_TYPE2OBJ[self.db_type]()

            env_var_helper = EnvironmentVariableHelper(
                env_file_path=self.credential.get_dot_env_file()
            )

            for cred_var, env_var in self.credential.VAR2ENV.items():
                if env_var_helper.set_name(env_var).exists():
                    self.set_credential_var(cred_var, env_var_helper.get_value())
                else:
                    from_file = env_var_helper.get_value_from_env_file()

                    if from_file:
                        self.set_credential_var(cred_var, from_file)
                    elif not ignore_cli:
                        self.set_credential_var(
                            cred_var,
                            self.ask_for_info(
                                cred_var,
                                getattr(self.credential, f"STD_{cred_var.upper()}"),
                            ),
                        )
                    else:
                        self.set_credential_var(
                            cred_var,
                            getattr(self.credential, f"STD_{cred_var.upper()}"),
                        )

                    env_var_helper.set_value_in_env_file(
                        str(getattr(self.credential, cred_var))
                    )

        return self

    @execute_if_authorized("")
    @start_if_not_started
    def get_uri(self) -> str:
        """
        Provides the URI to use.
        """

        return self.credential.get_uri()



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the list helpers

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import copy
from typing import Any, List, Optional


class ListHelper:
    """
    Simplify the list manipulation.

    :param subject:
        The list to work with.
    :param bool remove_empty: Process the deletion of empty strings.
    """

    _subject: Optional[List[Any]] = None

    def __init__(self, subject: Optional[List[Any]] = None):
        if subject is not None:
            self.subject = subject

    @property
    def subject(self):
        """
        Provides the current state of the :code:`_subject` attribute.
        """

        return self._subject

    @subject.setter
    def subject(self, value: List[Any]) -> None:
        """
        Sets the subject to work with.

        :param value:
            The subject to work with.

        :raise TypeError:
            When :code:`value` is not a :py:class:`list`.
        """

        if not isinstance(value, list):
            raise TypeError(f"<value> should be {list}, {type(value)} given.")

        self._subject = copy.deepcopy(value)

    def set_subject(self, value: List[Any]) -> "ListHelper":
        """
        Sets the subject to work with.

        :param value:
            The subject to work with.
        """

        self.subject = value

        return self

    def remove_empty(self) -> "ListHelper":
        """
        Removes the empty entries of the given list.
        """

        self.subject = [x for x in self.subject if x is None or x]

        return self

    def remove_duplicates(self) -> "ListHelper":
        """
        Removes the duplicates of the current list.
        """

        result = []

        for element in self.subject:
            if element not in result:
                result.append(element)

        self.subject = result

        return self

    def sort(self, *, reverse: bool = False) -> "ListHelper":
        """
        Sorts the given list (of string preferably).

         :param bool reverse: Tell us if we have to reverse the list.
        """

        self.custom_sort(str.lower, reverse=reverse)

        return self

    def custom_sort(self, key_method: Any, *, reverse: bool = False) -> "ListHelper":
        """
        Sorts the list with the given key method.

        :param key_method:
            A function or method to use to format the
            readed element before sorting.
        :type key_method: function|method

        :param bool reverse: Tell us if we have to reverse the list.
        """

        self.subject = sorted(self.subject, key=key_method, reverse=reverse)

        return self



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the endpoints related to the cleanup of the data we produce.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import argparse
import os
import sys
import traceback

import colorama

import PyFunceble.cli.facility
import PyFunceble.cli.storage
import PyFunceble.cli.utils.ascii_logo
import PyFunceble.facility
from PyFunceble.cli.filesystem.cleanup import FilesystemCleanup
from PyFunceble.cli.system.integrator import SystemIntegrator
from PyFunceble.cli.utils.testing import get_destination_from_origin
from PyFunceble.helpers.directory import DirectoryHelper


def cleaner() -> None:
    """
    Provides the CLI for the public file generation.
    """

    PyFunceble.facility.ConfigLoader.start()

    colorama.init(autoreset=True)

    description = (
        f"{colorama.Style.BRIGHT}{colorama.Fore.GREEN}PyFunceble Cleaner"
        f"{colorama.Style.RESET_ALL} - "
        "The cleaner of PyFunceble."
    )

    parser = argparse.ArgumentParser(
        description=description,
        epilog=PyFunceble.cli.storage.STD_EPILOG,
        add_help=True,
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument(
        "-f", "--file", type=str, help="Sets the file to cleanup the information for."
    )

    parser.add_argument(
        "-a",
        "--all",
        action="store_true",
        help="Authorizes the cleanup of everything that is not useful anymore.",
        default=False,
    )

    args = parser.parse_args()

    SystemIntegrator(args).start()

    directories_to_ignore = ["__pyfunceble_origin__"]

    utility = FilesystemCleanup()

    print(PyFunceble.cli.utils.ascii_logo.get_home_representation())

    to_cleanup = []
    directory_helper = DirectoryHelper()

    if args.file:
        to_cleanup.append(get_destination_from_origin(args.file))

    if (
        args.all
        and directory_helper.set_path(PyFunceble.cli.storage.OUTPUT_DIRECTORY).exists()
    ):
        to_cleanup.extend(
            [
                x
                for x in os.listdir(PyFunceble.cli.storage.OUTPUT_DIRECTORY)
                if directory_helper.set_path(
                    os.path.join(PyFunceble.cli.storage.OUTPUT_DIRECTORY, x)
                ).exists()
            ]
        )

    for directory in to_cleanup:
        if directory in directories_to_ignore:
            continue

        utility.set_parent_dirname(directory)

        print(f"Started cleanup of {utility.get_output_basedir()}.", end=" ")

        try:
            utility.clean_output_files()
            print(PyFunceble.cli.storage.DONE)
        except:  # pylint: disable=bare-except
            print(PyFunceble.cli.storage.ERROR)
            print(traceback.format_exc())
            sys.exit(1)

    if args.all:
        utility.clean_database()



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the base of all our downloader.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import datetime
import os
from typing import Optional

import PyFunceble.downloader.exceptions
import PyFunceble.exceptions
import PyFunceble.storage
from PyFunceble.helpers.dict import DictHelper
from PyFunceble.helpers.download import DownloadHelper
from PyFunceble.helpers.file import FileHelper


class DownloaderBase:
    """
    Provides the base of all downloader classes.

    The interface is actually simple, but the part which may be hard to
    understand is the "downtime" part. What we do, is that we save the
    download time inside a JSON file, so this class provides the base around
    the download mechanism but also the generation or update of that JSON file.
    """

    DOWNTIME_INDEX: Optional[str] = None
    """
    Used to set/track the download time of the current file.
    """

    DOWNLOAD_FREQUENCY: int = 1
    """
    The download frequency (in day).

    Example:
        if 1 is given, it's once every 24 hours.

    .. warning::
        A frequency of :code:`0` or a negative number will force the download
        every hour.
    """

    all_downtimes: Optional[dict] = {}
    """
    Stores the download time of all files (self managed).
    """

    _destination: Optional[str] = None
    _download_link: Optional[str] = None

    dict_helper: DictHelper = DictHelper()

    def __init__(self) -> None:
        self.downtimes_file = FileHelper(
            os.path.join(
                PyFunceble.storage.CONFIG_DIRECTORY, PyFunceble.storage.DOWN_FILENAME
            )
        )

        self.all_downtimes.update(self.get_all_downtimes())

    @property
    def authorized(self) -> bool:
        """
        Provides the authorization to start the download.
        """

        raise NotImplementedError()

    @property
    def destination(self) -> Optional[str]:
        """
        Provides the current state of the :code:`_destination` attribute.
        """

        return self._destination

    @destination.setter
    def destination(self, value: str) -> None:
        """
        Sets the destination.

        :param value:
            The value to set.

        :raise TypeError:
            When value is not a :py:class:`str`.
        """

        if not isinstance(value, str):
            raise TypeError(f"<value> should be {str}, {type(value)} given.")

        self._destination = value

    def set_destination(self, value: str) -> "DownloaderBase":
        """
        Sets the destination.

        :param value:
            The value to set.
        """

        self.destination = value

        return self

    @property
    def download_link(self) -> Optional[str]:
        """
        Provides the current state of the :code:`_download_link` attribute.
        """

        return self._download_link

    @download_link.setter
    def download_link(self, value: str) -> None:
        """
        Sets the link to download.

        :param value:
            The value to set.

        :raise TypeError:
            When value is not a :py:class:`str`.
        """

        if not isinstance(value, str):
            raise TypeError(f"<value> should be {str}, {type(value)} given.")

        self._download_link = value

    def set_download_link(self, value: str) -> "DownloaderBase":
        """
        Sets the link to download.

        :param value:
            The value to set.
        """

        self.download_link = value

        return self

    def get_all_downtimes(self) -> dict:
        """
        Provides the downloadtime of all files.
        """

        return self.dict_helper.from_json_file(self.downtimes_file.path)

    def is_downtime_set(self) -> bool:
        """
        Checks if the download time of the current object exists.
        """

        return (
            self.DOWNTIME_INDEX in self.all_downtimes
            and self.all_downtimes[self.DOWNTIME_INDEX]
            and all(
                x in self.all_downtimes[self.DOWNTIME_INDEX]
                and self.all_downtimes[self.DOWNTIME_INDEX][x]
                for x in ["iso", "timestamp"]
            )
        )

    def get_current_downtime(self):
        """
        Provides the download times of the current :code:`DOWN_TIME_INDEX`.
        """

        if self.is_downtime_set():
            return self.all_downtimes[self.DOWNTIME_INDEX]
        return None

    def set_current_downtime(self) -> "DownloaderBase":
        """
        Sets the current datetime into our registry.
        """

        current_datetime = datetime.datetime.utcnow()

        self.all_downtimes[self.DOWNTIME_INDEX] = {
            "iso": current_datetime.isoformat(),
            "timestamp": current_datetime.timestamp(),
        }

        return self

    def save_all_downtimes(self) -> None:
        """
        Saves the current state of the all downtimes.
        """

        self.dict_helper.set_subject(self.all_downtimes).to_json_file(
            self.downtimes_file.path
        )

    def is_last_download_expired(self) -> bool:
        """
        Checks if the last downloaded file is expired (if exists).
        """

        if not FileHelper(self.destination).exists() or not self.is_downtime_set():
            return True

        last_downloaded_time = datetime.datetime.fromtimestamp(
            self.get_current_downtime()["timestamp"]
        )

        if (
            self.DOWNLOAD_FREQUENCY <= 0
            and (datetime.datetime.utcnow() - last_downloaded_time).seconds < 3600
        ):
            return False

        if (
            last_downloaded_time + datetime.timedelta(days=self.DOWNLOAD_FREQUENCY)
            <= datetime.datetime.utcnow()
        ):
            return True

        return False

    def start(self) -> None:
        """
        Starts the download process.
        """

        if self.authorized and self.is_last_download_expired():
            if not hasattr(self, "destination") or not self.destination:
                raise PyFunceble.downloader.exceptions.NoDownloadDestinationGiven()

            if not hasattr(self, "download_link") or not self.download_link:
                raise PyFunceble.downloader.exceptions.NoDownloadLinkGiven()

            if DownloadHelper(self.download_link).download_text(
                destination=self.destination
            ):
                self.set_current_downtime()
                self.save_all_downtimes()



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides a way to convert/extract the network location of a given URL.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import urllib.parse
from typing import Any

from PyFunceble.converter.base import ConverterBase


class Url2Netloc(ConverterBase):
    """
    Provides the interface for the conversion/extration of the network location
    of a given URL.
    """

    @ConverterBase.data_to_convert.setter
    def data_to_convert(self, value: Any) -> None:
        """
        Overrites the default behavior.

        :raise TypeError:
            When the given data to convert is not :py:class:`str`
        :raise ValueError:
            When the given data to convert is empty.
        """

        if not isinstance(value, str):
            raise TypeError(f"<value> should be {str}, {type(value)} given.")

        if not value:
            raise ValueError("<value> should not be empty.")

        # pylint: disable=no-member
        super(Url2Netloc, self.__class__).data_to_convert.fset(self, value)

    def get_converted(self) -> str:
        """
        Provides the converted data (after conversion)
        """

        parsed_url = urllib.parse.urlparse(self.data_to_convert)

        if not parsed_url.netloc and parsed_url.path:
            netloc = parsed_url.path
        elif parsed_url.netloc:
            netloc = parsed_url.netloc
        else:  # pragma: no cover ## Safety
            netloc = self.data_to_convert

        if "//" in netloc:
            netloc = netloc[netloc.find("//") + 2 :]

        if "/" in netloc:
            netloc = netloc[: netloc.find("/")]

        return netloc



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the endpoints related to our production preparation.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import argparse
import sys
import traceback

import colorama

import PyFunceble.cli.storage
import PyFunceble.cli.utils.ascii_logo
from PyFunceble.cli.scripts.production import ProductionPrep


def producer() -> None:  # pylint: disable=too-many-statements
    """
    Provides the CLI for the production preparator.
    """

    colorama.init(autoreset=True)

    description = (
        f"{colorama.Style.BRIGHT}{colorama.Fore.GREEN}PyFunceble Production Preparator"
        f"{colorama.Style.RESET_ALL} - "
        "The Production peparator for PyFunceble."
    )

    parser = argparse.ArgumentParser(
        description=description,
        epilog=PyFunceble.cli.storage.STD_EPILOG,
        add_help=True,
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument(
        "branch", type=str, help="Sets the branch to acts as.", default=None
    )

    args = parser.parse_args()

    utility = ProductionPrep()

    try:
        utility.branch = args.branch
    except (TypeError, ValueError) as exception:
        print(f"{colorama.Fore.RED}{colorama.Style.BRIGHT}{str(exception)}")
        sys.exit(1)

    print(PyFunceble.cli.utils.ascii_logo.get_home_representation())

    print("Update setup.py", end=" ")
    try:
        utility.update_setup_py()
        print(PyFunceble.cli.storage.DONE)
    except:  # pylint: disable=bare-except
        print(PyFunceble.cli.storage.ERROR)
        print(traceback.format_exc())
        sys.exit(1)

    print("Update URL in source code and tests", end=" ")
    try:
        utility.update_code_urls()
        print(PyFunceble.cli.storage.DONE)
    except:  # pylint: disable=bare-except
        print(PyFunceble.cli.storage.ERROR)
        print(traceback.format_exc())
        sys.exit(1)

    print("Update URL in documentation", end=" ")
    try:
        utility.update_docs_urls()
        print(PyFunceble.cli.storage.DONE)
    except:  # pylint: disable=bare-except
        print(PyFunceble.cli.storage.ERROR)
        print(traceback.format_exc())
        sys.exit(1)

    print("Update source code and tests format", end=" ")
    try:
        utility.update_code_format()
        print(PyFunceble.cli.storage.DONE)
    except:  # pylint: disable=bare-except
        print(PyFunceble.cli.storage.ERROR)
        print(traceback.format_exc())
        sys.exit(1)

    print("Update version file", end=" ")
    try:
        utility.update_version_file()
        print(PyFunceble.cli.storage.DONE)
    except:  # pylint: disable=bare-except
        print(PyFunceble.cli.storage.ERROR)
        print(traceback.format_exc())
        sys.exit(1)

    print("Update dir structure file", end=" ")
    try:
        utility.update_dir_structure_file()
        print(PyFunceble.cli.storage.DONE)
    except:  # pylint: disable=bare-except
        print(PyFunceble.cli.storage.ERROR)
        print(traceback.format_exc())
        sys.exit(1)

    print("Update documentation", end=" ")
    try:
        utility.update_documentation()
        print(PyFunceble.cli.storage.DONE)
    except:  # pylint: disable=bare-except
        print(PyFunceble.cli.storage.ERROR)
        print(traceback.format_exc())
        sys.exit(1)



"""Fix some columns

Revision ID: 35c79626ecb9
Revises: ade87195b0a0
Create Date: 2020-08-21 11:42:07.044762

"""
from alembic import op
from sqlalchemy.exc import OperationalError

# pylint: skip-file

# revision identifiers, used by Alembic.
revision = "35c79626ecb9"
down_revision = "ade87195b0a0"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    try:
        op.create_index(
            op.f("ix_pyfunceble_mined_mined"),
            "pyfunceble_mined",
            ["mined"],
            unique=False,
        )
    except OperationalError:
        pass

    try:
        op.drop_index("mined", table_name="pyfunceble_mined")
    except OperationalError:
        pass

    try:
        op.create_index(
            op.f("ix_pyfunceble_status_tested"),
            "pyfunceble_status",
            ["tested"],
            unique=False,
        )
    except OperationalError:
        pass

    try:
        op.drop_index("tested", table_name="pyfunceble_status")
    except OperationalError:
        pass
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    try:
        op.create_index("tested", "pyfunceble_status", ["tested"], unique=True)
    except OperationalError:
        pass

    try:
        op.drop_index(
            op.f("ix_pyfunceble_status_tested"), table_name="pyfunceble_status"
        )
    except OperationalError:
        pass

    try:
        op.create_index("mined", "pyfunceble_mined", ["mined"], unique=True)
    except OperationalError:
        pass

    try:
        op.drop_index(op.f("ix_pyfunceble_mined_mined"), table_name="pyfunceble_mined")
    except OperationalError:
        pass
    # ### end Alembic commands ###



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides ans interface which let us interact with the Collection API.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import json
import logging
from datetime import datetime
from typing import List, Optional, Union

import requests
import requests.exceptions

import PyFunceble.facility
import PyFunceble.storage
from PyFunceble.checker.availability.status import AvailabilityCheckerStatus
from PyFunceble.checker.reputation.status import ReputationCheckerStatus
from PyFunceble.checker.syntax.status import SyntaxCheckerStatus
from PyFunceble.helpers.environment_variable import EnvironmentVariableHelper


class CollectionQueryTool:
    """
    Provides the interface to the collection dataset.

    :param token:
        The token to use to communicate with the API.

        .. warning::
            If :code:`None` is given, the class constructor will try to load the
            PYFUNCEBLE_COLLECTION_API_TOKEN environment variable.

    :param url_base:
        The base of the URL to communicate with.

    :param preferred_status_origin:
        The preferred data origin.
        It can be :code:`frequent`, :code:`latest` or :code:`recommended`.
    """

    SUPPORTED_CHECKERS: List[str] = ["syntax", "reputation", "availability"]
    SUPPORTED_STATUS_ORIGIN: List[str] = ["frequent", "latest", "recommended"]

    STD_URL_BASE: str = "http://localhost:8001"
    STD_PREFERRED_STATUS_ORIGIN: str = "frequent"

    _token: Optional[str] = None
    """
    The token to use while communicating with the collection API.
    """

    _url_base: Optional[str] = None
    """
    The base of the URL to communicate with.
    """

    _preferred_status_origin: Optional[str] = None
    """
    The preferred data origin
    """

    session: Optional[requests.Session] = None

    def __init__(
        self,
        *,
        token: Optional[str] = None,
        url_base: Optional[str] = None,
        preferred_status_origin: Optional[str] = None,
    ) -> None:

        if token is not None:
            self.token = token
        else:
            self.token = EnvironmentVariableHelper(
                "PYFUNCEBLE_COLLECTION_API_TOKEN"
            ).get_value(default="")

        if url_base is not None:
            self.url_base = url_base
        else:
            self.guess_and_set_url_base()

        if preferred_status_origin is not None:
            self.preferred_status_origin = preferred_status_origin
        else:
            self.guess_and_set_preferred_status_origin()

        self.session = requests.Session()
        self.session.headers.update(
            {
                "Authorization": f"Bearer {self.token}" if self.token else None,
                "Content-Type": "application/json",
            }
        )

    def __contains__(self, value: str) -> bool:
        """
        Checks if the given value is in the collection.

        :param value:
            The value to check.
        """

        return self.pull(value) is not None

    def __getitem__(self, value: str) -> Optional[dict]:
        """
        Gets the information about the given value.

        :param value:
            The value to get the information about.
        """

        return self.pull(value)

    @property
    def token(self) -> Optional[str]:
        """
        Provides the currently set token.
        """

        return self._token

    @token.setter
    def token(self, value: str) -> None:
        """
        Sets the value of the :code:`_token` attribute.

        :param value:
            The value to set.

        :raise TypeError:
            When the given :code:`value` is not a :py:class:`str`
        """

        if not isinstance(value, str):
            raise TypeError(f"<value> should be {str}, {type(value)} given.")

        self._token = value

    def set_token(self, value: str) -> "CollectionQueryTool":
        """
        Sets the value of the :code:`_token` attribute.

        :param value:
            The value to set.
        """

        self.token = value

        return self

    @property
    def url_base(self) -> Optional[str]:
        """
        Provides the value of the :code:`_url_base` attribute.
        """

        return self._url_base

    @url_base.setter
    def url_base(self, value: str) -> None:
        """
        Sets the base of the URL to work with.

        :param value:
            The value to set.

        :raise TypeError:
            When the given :code:`value` is not a :py:class:`str`.

        :raise ValueError:
            When the given :code:`value` does not have a scheme.
        """

        if not isinstance(value, str):
            raise TypeError(f"<value> should be {str}, {type(value)} given.")

        if not value.startswith(("http", "https")):
            raise ValueError(
                f"<value> is missing the scheme (http/https), {value} given."
            )

        self._url_base = value.rstrip("/")

    def set_url_base(self, value: str) -> "CollectionQueryTool":
        """
        Sets the base of the URL to work with.

        :parma value:
            The value to set.
        """

        self.url_base = value

        return self

    def guess_and_set_url_base(self) -> "CollectionQueryTool":
        """
        Try to guess the URL base to work with.
        """

        if PyFunceble.facility.ConfigLoader.is_already_loaded():
            if isinstance(PyFunceble.storage.CONFIGURATION.collection.url_base, str):
                self.url_base = PyFunceble.storage.CONFIGURATION.collection.url_base
            else:
                self.url_base = self.STD_URL_BASE
        else:
            self.url_base = self.STD_URL_BASE

        return self

    @property
    def preferred_status_origin(self) -> Optional[str]:
        """
        Provides the value of the :code:`_preferred_status_origin` attribute.
        """

        return self._preferred_status_origin

    @preferred_status_origin.setter
    def preferred_status_origin(self, value: str) -> None:
        """
        Sets the preferred status origin.

        :param value:
            The value to set.

        :raise TypeError:
            When the given :code:`value` is not a :py:class:`str`.

        :raise ValueError:
            When the given :code:`value` is not supported.
        """

        if not isinstance(value, str):
            raise TypeError(f"<value> should be {str}, {type(value)} given.")

        if value not in self.SUPPORTED_STATUS_ORIGIN:
            raise ValueError(f"<value> ({value}) is not supported.")

        self._preferred_status_origin = value

    def set_preferred_status_origin(self, value: str) -> "CollectionQueryTool":
        """
        Sets the preferred status origin.

        :parma value:
            The value to set.
        """

        self.preferred_status_origin = value

        return self

    def guess_and_set_preferred_status_origin(self) -> "CollectionQueryTool":
        """
        Try to guess the preferred status origin.
        """

        if PyFunceble.facility.ConfigLoader.is_already_loaded():
            if isinstance(
                PyFunceble.storage.CONFIGURATION.collection.preferred_status_origin, str
            ):
                self.preferred_status_origin = (
                    PyFunceble.storage.CONFIGURATION.collection.preferred_status_origin
                )
            else:
                self.preferred_status_origin = self.STD_PREFERRED_STATUS_ORIGIN
        else:
            self.preferred_status_origin = self.STD_PREFERRED_STATUS_ORIGIN

        return self

    def pull(self, subject: str) -> Optional[dict]:
        """
        Pulls all data related to the subject or :py:class:`None`

        :param subject:
            The subject to search for.

        :raise TypeError:
            When the given :code:`subject` is not a :py:class:`str`.

        :return:
            The response of the search.
        """

        logging.info("Starting to search subject: %r", subject)

        if not isinstance(subject, str):
            raise TypeError(f"<subject> should be {str}, {type(subject)} given.")

        url = f"{self.url_base}/v1/subject/search"

        try:
            response = self.session.post(
                url,
                json={"subject": subject},
            )

            response_json = response.json()

            if response.status_code == 200:
                logging.debug(
                    "Successfully search subject: %r. Response: %r",
                    subject,
                    response_json,
                )

                logging.info("Finished to search subject: %r", subject)

                return response_json
        except (requests.RequestException, json.decoder.JSONDecodeError):
            response_json = {}

        logging.debug(
            "Failed to search subject: %r. Response: %r", subject, response_json
        )
        logging.info("Finished to search subject: %r", subject)

        return None

    def push(
        self,
        checker_status: Union[
            AvailabilityCheckerStatus, SyntaxCheckerStatus, ReputationCheckerStatus
        ],
    ) -> Optional[dict]:
        """
        Push the given status to the collection.

        :param checker_status:
            The status to push.

        :raise TypeError:
            - When the given :code:`checker_status` is not a
              :py:class:`AvailabilityCheckerStatus`,
              :py:class:`SyntaxCheckerStatus` or
              :py:class:`ReputationCheckerStatus`.

            - When the given :code:`checker_status.subject` is not a
              :py:class:`str`.

        :raise ValueError:
            When the given :code:`checker_status.subject` is empty.
        """

        if not isinstance(
            checker_status,
            (AvailabilityCheckerStatus, SyntaxCheckerStatus, ReputationCheckerStatus),
        ):
            raise TypeError(
                f"<checker_status> should be {AvailabilityCheckerStatus}, "
                f"{SyntaxCheckerStatus} or {ReputationCheckerStatus}, "
                f"{type(checker_status)} given."
            )

        if not isinstance(checker_status.subject, str):
            raise TypeError(
                f"<checker_status.subject> should be {str}, "
                f"{type(checker_status.subject)} given."
            )

        if not isinstance(checker_status.checker_type, str):
            raise TypeError(
                f"<checker_status_checker_type> should be {str}, "
                f"{type(checker_status.subject)} given."
            )

        if checker_status.subject == "":
            raise ValueError("<checker_status.subject> cannot be empty.")

        status_to_subject = {
            "status": checker_status.status,
            "status_source": checker_status.status_source,
            "tested_at": checker_status.tested_at.isoformat(),
            "subject": checker_status.idna_subject,
        }

        if (
            hasattr(checker_status, "expiration_date")
            and checker_status.expiration_date
        ):
            self.__push_whois(
                {
                    "subject": checker_status.idna_subject,
                    "expiration_date": datetime.strptime(
                        checker_status.expiration_date, "%d-%b-%Y"
                    ).isoformat(),
                    "registrar": checker_status.registrar,
                }
            )

        data = self.__push_status(
            checker_status.checker_type.lower(), status_to_subject
        )

        return data

    def guess_all_settings(
        self,
    ) -> "CollectionQueryTool":  # pragma: no cover ## Underlying tested
        """
        Try to guess all settings.
        """

        to_ignore = ["guess_all_settings"]

        for method in dir(self):
            if method in to_ignore or not method.startswith("guess_"):
                continue

            getattr(self, method)()

        return self

    def __push_status(self, checker_type: str, data: dict) -> Optional[dict]:
        """
        Submits the given status to the collection.

        :param checker_type:
            The type of the checker.
        :param data:
            The data to submit.

        :raise TypeError:
            - When :code:`checker_type` is not a :py:class:`str`.

            - When :code:`data` is not a :py:class:`dict`.

        :raise ValueError:
            When the given :code:`checker_type` is not a subject checker type.
        """

        if not self.token:
            return None

        if checker_type not in self.SUPPORTED_CHECKERS:
            raise ValueError(f"<checker_type> ({checker_type}) is not supported.")

        logging.info("Starting to submit status: %r", data)

        url = f"{self.url_base}/v1/status/{checker_type}"

        try:
            response = self.session.post(
                url,
                json=data,
            )

            response_json = response.json()

            if response.status_code == 200:
                logging.debug("Successfully submitted data: %r to %s", data, url)

                logging.info("Finished to submit status: %r", data)
                return response_json
        except (requests.RequestException, json.decoder.JSONDecodeError):
            response_json = {}

        logging.debug(
            "Failed to submit data: %r to %s. Response: %r", data, url, response_json
        )

        logging.info("Finished to submit status: %r", data)

        return None

    def __push_whois(self, data: dict) -> Optional[dict]:
        """
        Submits the given WHOIS data into the given subject.

        :param checker_type:
            The type of the checker.
        :param data:
            The data to submit.

        :raise TypeError:
            - When :code:`checker_type` is not a :py:class:`str`.

            - When :code:`data` is not a :py:class:`dict`.

        :raise ValueError:
            When the given :code:`checker_type` is not a subject checker type.
        """

        if not self.token:
            return None

        if not isinstance(data, dict):  # pragma: no cover ## Should never happen
            raise TypeError(f"<data> should be {dict}, {type(data)} given.")

        logging.info("Starting to submit WHOIS: %r", data)

        url = f"{self.url_base}/v1/status/whois"

        try:
            response = self.session.post(
                url,
                json=data,
            )

            response_json = response.json()

            if response.status_code == 200:
                logging.debug("Successfully submitted WHOIS data: %r to %s", data, url)

                logging.info("Finished to submit WHOIS: %r", data)
                return response_json
        except (requests.RequestException, json.decoder.JSONDecodeError):
            response_json = {}

        logging.debug(
            "Failed to WHOIS data: %r to %s. Response: %r", data, url, response_json
        )

        logging.info("Finished to submit WHOIS: %r", data)
        return None



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the base of all our adapter.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from typing import Optional

import requests.adapters
import requests.models

import PyFunceble.storage
from PyFunceble.checker.syntax.ip import IPSyntaxChecker
from PyFunceble.query.dns.query_tool import DNSQueryTool


class RequestAdapterBase(requests.adapters.HTTPAdapter):
    """
    Extends the built-in HTTP adapater and acts as a base for all our own
    adapter.
    """

    resolving_cache: dict = {}
    resolving_use_cache: bool = False
    timeout: float = 5.0
    proxy_pattern: dict = {}

    def __init__(self, *args, **kwargs):
        if "timeout" in kwargs:
            self.timeout = float(kwargs["timeout"])
            del kwargs["timeout"]

        if "max_retries" in kwargs:
            kwargs["max_retries"] = requests.adapters.Retry(
                total=kwargs["max_retries"], respect_retry_after_header=False
            )

        if "dns_query_tool" in kwargs:
            self.dns_query_tool = kwargs["dns_query_tool"]
            del kwargs["dns_query_tool"]
        else:
            self.dns_query_tool = DNSQueryTool()

        if "proxy_pattern" in kwargs:
            self.proxy_pattern = kwargs["proxy_pattern"]
            del kwargs["proxy_pattern"]
        else:
            self.proxy_pattern = {}

        super().__init__(*args, **kwargs)

    @staticmethod
    def fake_response() -> requests.models.Response:
        """
        Provides the fake response that is provided when we couldn't resolve the
        given domain.
        """

        raise PyFunceble.factory.Requester.exceptions.ConnectionError(
            "Could not resolve."
        )

    @staticmethod
    def extract_extension(subject: str) -> Optional[str]:
        """
        Provides the extension of the given subject.

        .. versionchanged:: 4.1.1.dev
            Handle the case that the given subject does not have a `.` (point).

        :param str subject:
            The subject to get extract the extension from.

        :raise TypeError:
            When the given :code:`subject` is not a :py:class:`str`.
        :raise ValueError:
            When the given :code:`subject` is an empty :py:class:`str`.
        """

        if not isinstance(subject, str):
            raise TypeError(f"<subject> should be {str}, type(subject) given.")

        if not subject:
            raise ValueError("<subject> should not be empty.")

        if "." not in subject:
            return None

        if subject.endswith("."):
            # Absolute needs a little correction.
            last_point = subject[:-1].rfind(".")
        else:
            last_point = subject.rindex(".")

        extension = subject[last_point + 1 :]

        if extension.endswith("."):
            return extension[:-1]
        return extension

    def fetch_proxy_from_pattern(self, subject: str) -> dict:
        """
        Provides the proxy settings to use for the given subject.

        .. versionchanged:: 4.1.1.dev
            Handle the case that the given subject has no extension/TLD.

        :param str subject:
            The subject to work with.

        :raise TypeError:
            When the given :code:`subject` is not a :py:class:`str`.
        :raise ValueError:
            When the given :code:`subject` is an empty :py:class:`str`.
        """

        def correct_input(pattern_input: dict) -> dict:
            result = {}

            if "http" in pattern_input and pattern_input["http"]:
                result["http"] = pattern_input["http"]

            if "https" in pattern_input and pattern_input["https"]:
                result["https"] = pattern_input["https"]

            if "http" in result and "https" not in result:
                result["https"] = result["http"]

            if "https" in result and "http" not in result:
                result["http"] = result["https"]

            return result

        if not isinstance(subject, str):
            raise TypeError(f"<subject> should be {str}, type(subject) given.")

        if not subject:
            raise ValueError("<subject> should not be empty.")

        extension = self.extract_extension(subject)

        proxies = {}

        if extension and "rules" in self.proxy_pattern:
            for rule in self.proxy_pattern["rules"]:
                local_proxy = {}

                if "http" in rule and rule["http"]:
                    local_proxy["http"] = rule["http"]
                if "https" in rule and rule["https"]:
                    local_proxy["https"] = rule["https"]

                if not local_proxy:
                    continue

                if "tld" in rule and extension in rule["tld"]:
                    proxies = correct_input(local_proxy)
                    break

        if not proxies and "global" in self.proxy_pattern:
            proxies = correct_input(self.proxy_pattern["global"])

        return proxies

    def resolve_with_cache(self, hostname: str) -> Optional[str]:
        """
        Try to resolve using an internal cache.
        """

        if hostname not in self.resolving_cache:
            self.resolving_cache[hostname] = self.resolve_without_cache(hostname)

        return self.resolving_cache[hostname]

    def resolve_without_cache(self, hostname: str) -> Optional[str]:
        """
        Resolves the IP of the given hostname.

        :param hostname:
            The hostname to get resolve.
        """

        def get_last_cname(subject: str, recursion_depth: int = 60) -> Optional[str]:
            """
            Given a subject, this function tries to query the CNAME until there
            is none.

            :param subject:
                The first subject.
            """

            last_cname_result = []
            last_cname_new_subject = subject

            depth = 0

            while depth < recursion_depth:
                local_last_cname_result = (
                    self.dns_query_tool.set_query_record_type("CNAME")
                    .set_subject(last_cname_new_subject)
                    .query()
                )

                depth += 1

                if any(x in last_cname_result for x in local_last_cname_result):
                    break

                last_cname_result.extend(local_last_cname_result)

                if local_last_cname_result:
                    last_cname_new_subject = local_last_cname_result[0]
                else:
                    break

            try:
                return last_cname_result[-1]
            except IndexError:
                return None

        result = set()

        if not IPSyntaxChecker(hostname).is_valid():
            last_cname = get_last_cname(hostname)

            if last_cname:
                result.update(
                    self.dns_query_tool.set_query_record_type("A")
                    .set_subject(last_cname)
                    .query()
                )
            else:
                result.update(
                    self.dns_query_tool.set_query_record_type("A")
                    .set_subject(hostname)
                    .query()
                )
        else:
            result.add(hostname)

        if result:
            return result.pop()
        return None

    def resolve(self, hostname: str) -> Optional[str]:
        """
        Resolves with the prefered method.
        """

        if hostname:
            if self.resolving_use_cache:
                return self.resolve_with_cache(hostname)
            return self.resolve_without_cache(hostname)
        return None



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the schema of our "whois_record" table.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from datetime import datetime

from sqlalchemy import Column, DateTime, String, Text

from PyFunceble.database.sqlalchemy.base_schema import SchemaBase


class Inactive(SchemaBase):
    """
    Provides the schema of our inactive table.
    """

    idna_subject = Column(Text, nullable=False)
    checker_type = Column(String(length=50), nullable=False)
    destination = Column(Text, nullable=False)
    tested_at = Column(DateTime(), default=datetime.utcnow, nullable=False)



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the CI engine and detection tool for the GitHub Actions.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import PyFunceble.facility
from PyFunceble.cli.continuous_integration.base import ContinuousIntegrationBase
from PyFunceble.helpers.environment_variable import EnvironmentVariableHelper


class GitHubActions(ContinuousIntegrationBase):
    """
    Provides the interface which detects and work under the GitHub Actions
    infrastructure.
    """

    end_commit_marker: str = "[GHA skip]"

    def guess_and_set_authorized(self) -> "GitHubActions":
        """
        Tries to guess the authorization.
        """

        needed_environment_vars = ["GITHUB_ACTIONS"]

        if PyFunceble.facility.ConfigLoader.is_already_loaded():
            if bool(PyFunceble.storage.CONFIGURATION.cli_testing.ci.active):
                self.authorized = all(
                    EnvironmentVariableHelper(x).exists()
                    for x in needed_environment_vars
                )
            else:
                super().guess_and_set_authorized()
        elif all(
            EnvironmentVariableHelper(x).exists() for x in needed_environment_vars
        ):
            self.authorized = True
        else:
            super().guess_and_set_authorized()

        return self

    def guess_and_set_token(self) -> "GitHubActions":
        """
        Tries to guess and set the token.
        """

        environment_var = EnvironmentVariableHelper("GITHUB_TOKEN")

        if environment_var.exists():
            self.token = environment_var.get_value()

        return self



"""Add registrar column

Revision ID: 912bbcb77a6c
Revises: 6f4729deaf03
Create Date: 2021-12-04 23:52:11.861732

"""
import sqlalchemy as sa
from alembic import op

# pylint: skip-file

# revision identifiers, used by Alembic.
revision = "912bbcb77a6c"
down_revision = "6f4729deaf03"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "pyfunceble_whois_record", sa.Column("registrar", sa.Text(), nullable=True)
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("pyfunceble_whois_record", "registrar")
    # ### end Alembic commands ###



# pylint:disable=line-too-long
"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the default input line converter.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from typing import Any, List

from PyFunceble.checker.syntax.ip import IPSyntaxChecker
from PyFunceble.converter.base import ConverterBase


class InputLine2Subject(ConverterBase):
    """
    Converts/Extract the subjcts to test from an inputed line.
    """

    COMMENT: str = "#"
    PARTICULAR_COMMENT: List[str] = ["!"]
    SPACE: str = " "
    NSLOOKUP_SPACE: str = "\\032"
    TAB: str = "\t"

    @ConverterBase.data_to_convert.setter
    def data_to_convert(self, value: Any) -> None:
        """
        Overrites the default behavior.

        :raise TypeError:
            When the given data to convert is not :py:class:`str`
        """

        if not isinstance(value, str):
            raise TypeError(f"<value> should be {str}, {type(value)} given.")

        # pylint: disable=no-member
        super(InputLine2Subject, self.__class__).data_to_convert.fset(self, value)

    def get_converted(self) -> List[str]:
        """
        Provides the subject to test.
        """

        result = []

        subject = self.data_to_convert.strip()

        if subject and (
            not subject.startswith(self.COMMENT)
            and any(not subject.startswith(x) for x in self.PARTICULAR_COMMENT)
        ):
            if self.COMMENT in subject:
                subject = subject[: subject.find(self.COMMENT)].strip()

            if self.NSLOOKUP_SPACE in subject:
                # Comply with RFC 6367:
                #    Note that nslookup escapes spaces as "\032" for display
                #    purposes, but a graphical DNS-SD browser should not.
                subject = subject.replace(self.NSLOOKUP_SPACE, self.SPACE)

            if self.SPACE in subject or self.TAB in subject:
                splitted = subject.split()

                if IPSyntaxChecker(splitted[0]).is_valid():
                    # This is for the hosts format.
                    # If the first entry is an IP, we will only extract
                    # the entries after the first one.
                    result.extend(splitted[1:])
                else:
                    # All other cases, we extract every entries.
                    result.extend(splitted)
            else:
                result.append(subject)
        return result



"""Add continue table.

Revision ID: 3a4c55a9320d
Revises: 3d6f4a33cdb2
Create Date: 2020-12-13 22:52:56.968513

"""
import sqlalchemy as sa
from alembic import op

# pylint: skip-file

# revision identifiers, used by Alembic.
revision = "3a4c55a9320d"
down_revision = "3d6f4a33cdb2"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "pyfunceble_continue",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("created_at", sa.DateTime(), nullable=False),
        sa.Column("modified_at", sa.DateTime(), nullable=True),
        sa.Column("idna_subject", sa.Text(), nullable=False),
        sa.Column("checker_type", sa.String(length=50), nullable=False),
        sa.Column("destination", sa.Text(), nullable=False),
        sa.Column("source", sa.Text(), nullable=False),
        sa.Column("tested_at", sa.DateTime(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("pyfunceble_continue")
    # ### end Alembic commands ###



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides some utilities related to the CLI stdout.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import secrets

import colorama

import PyFunceble.facility
import PyFunceble.storage


def get_template_to_use() -> str:
    """
    Provides the template to use.
    """

    if PyFunceble.storage.CONFIGURATION.cli_testing.display_mode.simple:
        return "simple"
    if PyFunceble.storage.CONFIGURATION.cli_testing.display_mode.all:
        return "all"

    return "less"


def print_single_line(value: str = ".", end: str = "", *, force: bool = False) -> None:
    """
    Prints the given :code:`value` in the current line.

    :param value:
        The default value
    :param end:
        Same as the end argument of the built-in print function.
    :param force:
        Forces the printing.
    """

    if PyFunceble.facility.ConfigLoader.is_already_loaded():
        if (
            force
            or PyFunceble.storage.CONFIGURATION.cli_testing.ci.active
            or PyFunceble.storage.CONFIGURATION.cli_testing.display_mode.dots
        ):
            print(value, end=end)


def print_thanks() -> None:
    """
    Randomly prints our thanks message.
    """

    if PyFunceble.facility.ConfigLoader.is_already_loaded():
        if (
            not PyFunceble.storage.CONFIGURATION.cli_testing.display_mode.quiet
            and not PyFunceble.storage.CONFIGURATION.cli_testing.display_mode.simple
        ):
            if PyFunceble.storage.CONFIGURATION.cli_testing.display_mode.colour:
                print(
                    f"\n{colorama.Fore.GREEN}{colorama.Style.BRIGHT}"
                    f"Thank you for using PyFunceble!{colorama.Style.RESET_ALL}"
                )
            else:
                print("\nThank you for using PyFunceble!")

            if int(secrets.token_hex(8), 16) % 3 == 0:
                if PyFunceble.storage.CONFIGURATION.cli_testing.display_mode.colour:
                    print(
                        f"{colorama.Fore.YELLOW}{colorama.Style.BRIGHT}"
                        f"Share your experience on {colorama.Fore.CYAN}Twitter "
                        f"{colorama.Fore.YELLOW}with {colorama.Fore.CYAN}"
                        f"#PyFunceble{colorama.Fore.YELLOW} "
                        f"or {colorama.Fore.CYAN}@PyFunceble"
                        f"{colorama.Fore.YELLOW}!"
                    )
                else:
                    print(
                        "Share your experience on Twitter with #PyFunceble or "
                        "@PyFunceble!"
                    )

            if int(secrets.token_hex(8), 16) % 3 == 0:
                if PyFunceble.storage.CONFIGURATION.cli_testing.display_mode.colour:
                    print(
                        f"{colorama.Fore.YELLOW}{colorama.Style.BRIGHT}"
                        f"Do you have a feedback, an issue or an improvement idea? "
                        f"{colorama.Fore.YELLOW}Let us know on {colorama.Fore.CYAN}"
                        f"GitHub{colorama.Fore.YELLOW}!"
                    )
                else:
                    print(
                        "Do you ave a feedback, an issue or an improvement idea? "
                        "Let us know on GitHub!"
                    )



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the base of all WHOIS related dataset.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from datetime import datetime
from typing import List, Union

from PyFunceble.database.sqlalchemy.all_schemas import WhoisRecord
from PyFunceble.dataset.db_base import DBDatasetBase


class WhoisDatasetBase(DBDatasetBase):
    """
    Provides the base of all Whois related interface.
    """

    FIELDS: List[str] = [
        "subject",
        "idna_subject",
        "expiration_date",
        "epoch",
        "registrar",
    ]

    COMPARISON_FIELDS: List[str] = ["subject", "idna_subject"]

    @staticmethod
    def is_expired(row: Union[dict, WhoisRecord]) -> bool:
        """
        Given a row, we look if the row is expired.
        """

        if isinstance(row, WhoisRecord):
            to_check = row.epoch
        elif "epoch" in row:
            to_check = row["epoch"]
        else:
            return True

        return datetime.utcnow() > datetime.fromtimestamp(float(to_check))

    @DBDatasetBase.execute_if_authorized(None)
    def get_filtered_row(self, row: Union[dict, WhoisRecord]) -> dict:
        """
        Removes all unkowns fields (not declared) from the given row.

        :param row:
            The row to work with.
        """

        if isinstance(row, WhoisDatasetBase):
            row = row.to_dict()

        result = super().get_filtered_row(row)

        if "epoch" in result and isinstance(result["epoch"], float):
            # We do this here because we have to convert to a string in
            # order to be able to write into the CSV file.
            result["epoch"] = str(result["epoch"])

        return result



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the base of all Mariadb stored datasets.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import functools
from datetime import datetime
from typing import Any, Generator, Optional

import sqlalchemy.exc
from sqlalchemy.orm import Session

import PyFunceble.cli.factory
import PyFunceble.sessions
from PyFunceble.dataset.db_base import DBDatasetBase


class MariaDBDatasetBase(DBDatasetBase):
    """
    Provides the base of all MariaDB stored dataset.
    """

    STD_KEEP_SESSION_OPEN: bool = False
    ORM_OBJ = None

    db_session: Optional[Session] = None

    def __init__(
        self,
        *,
        authorized: Optional[bool] = None,
        remove_unneeded_fields: Optional[bool] = None,
        db_session: Optional[Session] = None,
    ) -> None:
        self.db_session = db_session

        super().__init__(
            authorized=authorized, remove_unneeded_fields=remove_unneeded_fields
        )

    def __contains__(self, value: str) -> bool:
        raise NotImplementedError()

    def __getitem__(self, value: Any) -> Any:
        raise KeyError(value)

    def ensure_orm_obj_is_given(func):  # pylint: disable=no-self-argument
        """
        Ensures that the ORM object is given before launching the decorated
        method.

        :raise RuntimeError:
            When :code:`ORM_OBJ` is not declared.
        """

        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            if self.ORM_OBJ is None:
                raise RuntimeError("<self.ORM_OBJ> not given.")

            return func(self, *args, **kwargs)  # pylint: disable=not-callable

        return wrapper

    @DBDatasetBase.execute_if_authorized(None)
    def get_content(self) -> Generator[dict, None, None]:
        """
        Provides a generator which provides the next dataset to read.
        """

        for row in self.db_session.query(self.ORM_OBJ):
            row = row.to_dict()

            yield row

    @DBDatasetBase.execute_if_authorized(None)
    def update(self, row, *, ignore_if_exist: bool = False) -> "MariaDBDatasetBase":
        """
        Adds the given dataset into the database if it does not exists.
        Update otherwise.

        .. note::
            This should be the preferred method for introduction of new dataset.

        :raise TypeError:
            When the given :code:`row` is not a :py:class:`dict` or sqlalchemy
            schema.
        """

        if not isinstance(row, (dict, type(self.ORM_OBJ))):
            raise TypeError(
                f"<row> should be {dict} or {self.ORM_OBJ}, {type(row)} given."
            )

        PyFunceble.facility.Logger.info("Started to update row.")

        if isinstance(row, type(self.ORM_OBJ)):
            row = row.to_dict()

        if "id" in row:
            self.db_session.execute(
                self.ORM_OBJ.__table__.update().where(self.ORM_OBJ.id == row["id"]),
                row,
            )

            self.db_session.commit()
        else:
            existing_row_id = self.get_existing_row_id(row)

            if existing_row_id is not None:
                if not ignore_if_exist:
                    row["id"] = existing_row_id
                    self.update(row, ignore_if_exist=ignore_if_exist)
            else:
                self.add(row)

        PyFunceble.facility.Logger.debug("Updated row:\n%r", row)
        PyFunceble.facility.Logger.info("Finished to update row.")

        return self

    @DBDatasetBase.execute_if_authorized(None)
    @ensure_orm_obj_is_given
    def remove(self, row) -> "MariaDBDatasetBase":
        """
        Removes the given dataset from the database.

        :param row:
            The row or dataset to check.

        :raise TypeError:
            When the given :code:`row` is not a :py:class:`dict` or sqlalchemy
            schema.
        """

        if not isinstance(row, (dict, type(self.ORM_OBJ))):
            raise TypeError(
                f"<row> should be {dict} or {self.ORM_OBJ}, {type(row)} given."
            )

        PyFunceble.facility.Logger.info("Started to remove row.")

        if not isinstance(row, type(self.ORM_OBJ)):
            row_id = self.get_existing_row_id(row)

            if row_id is not None:
                self.db_session.query(self.ORM_OBJ).filter(
                    self.ORM_OBJ == row_id
                ).delete()
                self.db_session.commit()
        else:
            self.db_session.delete(row)
            self.db_session.commit()

        PyFunceble.facility.Logger.debug("Removed row:\n%r", row)
        PyFunceble.facility.Logger.info("Finished to remove row.")

        return self

    @DBDatasetBase.execute_if_authorized(False)
    @ensure_orm_obj_is_given
    def exists(self, row) -> bool:
        """
        Checks if the given dataset exists in our dataset.

        :param row:
            The row or dataset to check.

        :raise TypeError:
            When the given :code:`row` is not a :py:class:`dict` or sqlalchemy
            schema.
        """

        return self.get_existing_row_id(row) is not None

    @DBDatasetBase.execute_if_authorized(None)
    @ensure_orm_obj_is_given
    def get_existing_row_id(self, row):
        """
        Returns the ID of the existing row.

        :param row:
            The row or dataset to check,
        """

        if not isinstance(row, (dict, type(self.ORM_OBJ))):
            raise TypeError(
                f"<row> should be {dict} or {self.ORM_OBJ}, {type(row)} given."
            )

        if isinstance(row, type(self.ORM_OBJ)):
            row = row.to_dict()

        result = self.db_session.query(self.ORM_OBJ.id)

        for field in self.COMPARISON_FIELDS:
            result = result.filter(getattr(self.ORM_OBJ, field) == row[field])

        result = result.first()

        if result is not None:
            return result.id
        return None

    @DBDatasetBase.execute_if_authorized(None)
    @ensure_orm_obj_is_given
    def get_existing_row(self, row):
        """
        Returns the matching row.

        :param row:
            The row or dataset to check,
        """

        if not isinstance(row, (dict, type(self.ORM_OBJ))):
            raise TypeError(
                f"<row> should be {dict} or {self.ORM_OBJ}, {type(row)} given."
            )

        if isinstance(row, type(self.ORM_OBJ)):
            row = row.to_dict()

        result = self.db_session.query(self.ORM_OBJ)

        for field in self.COMPARISON_FIELDS:
            result = result.filter(getattr(self.ORM_OBJ, field) == row[field])

        return result.first()

    @DBDatasetBase.execute_if_authorized(None)
    @ensure_orm_obj_is_given
    def add(self, row) -> "MariaDBDatasetBase":
        """
        Adds the given dataset into the database.

        :param row:
            The row or dataset to add.

        :raise TypeError:
            When the given :code:`row` is not a :py:class:`dict` or sqlalchemy
            schema.
        """

        if not isinstance(row, (dict, type(self.ORM_OBJ))):
            raise TypeError(
                f"<row> should be {dict} or {self.ORM_OBJ}, {type(row)} given."
            )

        PyFunceble.facility.Logger.info("Started to add row.")

        if isinstance(row, type(self.ORM_OBJ)):
            row = row.to_dict()

        try:
            self.db_session.execute(self.ORM_OBJ.__table__.insert(), row)
            self.db_session.commit()
        except sqlalchemy.exc.DataError as exception:
            if "expiration_date" not in row and "epoch" not in row:
                raise exception

            y2k38_limit = datetime(2037, 12, 31, 0, 0)
            new_date = datetime.fromtimestamp(float(row["epoch"]))
            new_date -= new_date - y2k38_limit

            row["epoch"] = str(new_date.timestamp())
            row["expiration_date"] = new_date.strftime("%d-%b-%Y")

            self.db_session.execute(self.ORM_OBJ.__table__.insert(), row)
            self.db_session.commit()

        PyFunceble.facility.Logger.debug("Added row:\n%r", row)
        PyFunceble.facility.Logger.info("Finished to add row.")

        return self



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the base of all our credential holders.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import functools
import os
from typing import List, Optional

import PyFunceble.storage
from PyFunceble.helpers.file import FileHelper


class CredentialBase:
    """
    Provides the base of all our credential holder.
    """

    STD_HOST: str = "localhost"
    STD_PORT: int = 3306
    STD_NAME: str = PyFunceble.storage.PROJECT_NAME.lower()
    STD_USERNAME: str = PyFunceble.storage.PROJECT_NAME.lower()
    STD_PASSWORD: str = f"{PyFunceble.storage.PROJECT_NAME}:15_93le"
    STD_CHARSET: str = "utf8mb4"

    VAR2ENV: dict = {
        "host": "PYFUNCEBLE_DB_HOST",
        "port": "PYFUNCEBLE_DB_PORT",
        "name": "PYFUNCEBLE_DB_NAME",
        "username": "PYFUNCEBLE_DB_USERNAME",
        "password": "PYFUNCEBLE_DB_PASSWORD",
        "charset": "PYFUNCEBLE_DB_CHARSET",
    }
    """
    Maps our credential variable with environment variable.
    """

    dotenv_locations: List[str] = []
    """
    Provides the location of the dotenv to work with.

    .. warning::
        The order is important. The last one in the list will be taken as
        default if everything else is not found in the filesystem.
    """

    protocol: Optional[str] = None

    _host: Optional[str] = None
    _port: Optional[int] = None
    _name: Optional[str] = None
    _username: Optional[str] = None
    _password: Optional[str] = None
    _charset: Optional[str] = None

    def __init__(
        self,
        *,
        host: Optional[str] = None,
        port: Optional[int] = None,
        name: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        charset: Optional[str] = None,
    ) -> None:

        if host is not None:
            self.host = host
        else:
            self.host = self.STD_HOST

        if port is not None:
            self.port = port
        else:
            self.port = self.STD_PORT

        if name is not None:
            self.name = name
        else:
            self.name = self.STD_NAME

        if username is not None:
            self.username = username
        else:
            self.username = self.STD_USERNAME

        if password is not None:
            self.password = password
        else:
            self.password = self.STD_PASSWORD

        if charset is not None:
            self.charset = charset
        else:
            self.charset = self.STD_CHARSET

        self.dotenv_locations = [
            os.path.realpath(PyFunceble.storage.ENV_FILENAME),
            os.path.join(
                PyFunceble.storage.CONFIG_DIRECTORY, PyFunceble.storage.ENV_FILENAME
            ),
        ]

    def ensure_protocol_is_given(func):  # pylint: disable=no-self-argument
        """
        Ensures that the protocol is given before launching the decorated
        method.

        :raise ValueError:
            When the :code:`protocol` is not given.
        """

        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            if not self.protocol:
                raise ValueError("<self.protocol> is not given.")

            return func(self, *args, **kwargs)  # pylint: disable=not-callable

        return wrapper

    @property
    def host(self) -> Optional[str]:
        """
        Provides the current state of the :code:`_host` attribute.
        """

        return self._host

    @host.setter
    def host(self, value: str) -> None:
        """
        Sets the hosts to interact with.

        :param value:
            The value to set.

        :raise TypeError:
            When the given :code:`value` is not a :py:class:`str`.
        :raise ValueError:
            When the given :code:`value` is empty.
        """

        if not isinstance(value, str):
            raise TypeError(f"<value> should be {str}, {type(value)} given.")

        if not value:
            raise ValueError("<value> should not be empty.")

        self._host = value

    def set_host(self, value: str) -> "CredentialBase":
        """
        Sets the hosts to interact with.

        :param value:
            The value to set.
        """

        self.host = value

        return self

    @property
    def port(self) -> Optional[int]:
        """
        Provides the current state of the :code:`_port` attribute.
        """

        return self._port

    @port.setter
    def port(self, value: int) -> None:
        """
        Sets the port to interact with.

        :param value:
            The value to set.

        :raise TypeError:
            When the given :code:`value` is not a :py:class:`int`.
        """

        if not isinstance(value, int):
            raise TypeError(f"<value> should be {int}, {type(value)} given.")

        self._port = value

    def set_port(self, value: int) -> "CredentialBase":
        """
        Sets the port to interact with.

        :param value:
            The value to set.
        """

        self.port = value

        return self

    @property
    def name(self) -> Optional[str]:
        """
        Provides the current state of the :code:`_name` attribute.
        """

        return self._name

    @name.setter
    def name(self, value: str) -> None:
        """
        Sets the name of the database to interact with.

        :param value:
            The value to set.

        :raise TypeError:
            When the given :code:`value` is not a :py:class:`str`.
        :raise ValueError:
            When the given :code:`value` is empty.
        """

        if not isinstance(value, str):
            raise TypeError(f"<value> should be {str}, {type(value)} given.")

        if not value:
            raise ValueError("<value> should not be empty.")

        self._name = value

    def set_name(self, value: str) -> "CredentialBase":
        """
        Sets the name of the database to interact with.

        :param value:
            The value to set.
        """

        self.name = value

        return self

    @property
    def username(self) -> Optional[str]:
        """
        Provides the current state of the :code:`_username` attribute.
        """

        return self._username

    @username.setter
    def username(self, value: str) -> None:
        """
        Sets the username to use to authenticate ourselves.

        :param value:
            The value to set.

        :raise TypeError:
            When the given :code:`value` is not a :py:class:`str`.
        :raise ValueError:
            When the given :code:`value` is empty.
        """

        if not isinstance(value, str):
            raise TypeError(f"<value> should be {str}, {type(value)} given.")

        if not value:
            raise ValueError("<value> should not be empty.")

        self._username = value

    def set_username(self, value: str) -> "CredentialBase":
        """
        Sets the username to use to authenticate ourselves.

        :param value:
            The value to set.
        """

        self.username = value

        return self

    @property
    def password(self) -> Optional[str]:
        """
        Provides the current state of the :code:`_password` attribute.
        """

        return self._password

    @password.setter
    def password(self, value: str) -> None:
        """
        Sets the password to use to authenticate ourselves.

        :param value:
            The value to set.

        :raise TypeError:
            When the given :code:`value` is not a :py:class:`str`.
        """

        if not isinstance(value, str):
            raise TypeError(f"<value> should be {str}, {type(value)} given.")

        self._password = value

    def set_password(self, value: str) -> "CredentialBase":
        """
        Sets the password to use to authenticate ourselves.

        :param value:
            The value to set.
        """

        self.host = value

        return self

    @property
    def charset(self) -> Optional[str]:
        """
        Provides the current state of the :code:`_charset` attribute.
        """

        return self._charset

    @charset.setter
    def charset(self, value: str) -> None:
        """
        Sets the charset to use.

        :param value:
            The value to set.

        :raise TypeError:
            When the given :code:`value` is not a :py:class:`str`.
        :raise ValueError:
            When the given :code:`value` is empty.
        """

        if not isinstance(value, str):
            raise TypeError(f"<value> should be {str}, {type(value)} given.")

        if not value:
            raise ValueError("<value> should not be empty.")

        self._charset = value

    def set_charset(self, value: str) -> "CredentialBase":
        """
        Sets the charset to use.

        :param value:
            The value to set.
        """

        self.charset = value

        return self

    @ensure_protocol_is_given
    def get_uri(self) -> str:
        """
        Provides the SQLAlchemy URI.
        """

        if self.host.startswith("/"):
            return (
                f"{self.protocol}://{self.username}:{self.password}"
                f"@localhost/{self.name}?unix_socket={self.host}"
                f"&charset={self.charset}"
            )

        return (
            f"{self.protocol}://{self.username}:{self.password}"
            f"@{self.host}/{self.name}"
            f"?charset={self.charset}"
        )

    def get_dot_env_file(self) -> str:
        """
        Provides the dotenv file to work with.
        """

        file_helper = FileHelper()

        for file in self.dotenv_locations:
            if file_helper.set_path(file).exists():
                return file_helper.path

        return self.dotenv_locations[-1]
